var canvas = document.getElementById('canvas');
var context = canvas.getContext('2d');
var ast = null;

var editor = document.getElementById('editor');
editor.oninput = function() {
  try {
    var lexer = new Lexer(editor.value);
    var tokens = lexer.lex();
    var parser = new Parser(tokens);
    ast = parser.parse();
    execute();
    document.getElementById('console').innerHTML = '';
  } catch (e) {
    console.log(e);
    document.getElementById('console').innerHTML = e; 
  }
}

var scrubber = document.getElementById('scrubber');
scrubber.oninput = execute;

function execute() {
  context.clearRect(0, 0, canvas.width, canvas.height);
  var scope = {red: 0, green: 0, blue: 255, t: scrubber.value};
  ast.evaluate(scope);
}

// COMMON ---------------------------------------------------------------------

var INTEGER = 'INTEGER';
var IDENTIFIER = 'IDENTIFIER';
var EOF = 'EOF';
var LEFT_PARENTHESIS = 'LEFT_PARENTHESIS';
var RIGHT_PARENTHESIS = 'RIGHT_PARENTHESIS';
var PLUS = 'PLUS';
var TIMES = 'TIMES';
var ASSIGN = 'ASSIGN';
var MINUS = 'MINUS';
var DIVIDE = 'DIVIDE';
var RECTANGLE = 'RECTANGLE';
var CIRCLE = 'CIRCLE';
var REPEAT = 'REPEAT';
var END = 'END';
var TO = 'TO';

function Token(type, source) {
  this.type = type;
  this.source = source;
}

// LEXER ----------------------------------------------------------------------

function Lexer(program) {
  // private instance variables (var defs)
  var iToRead;
  var tokenSoFar;
  var tokens;

  function has(regex) {
    return regex.test(program.charAt(iToRead));
  }

  function consume() {
    tokenSoFar += program.charAt(iToRead);
    ++iToRead;
  }

  function emitToken(type) {
    tokens.push(new Token(type, tokenSoFar));
    tokenSoFar = '';
  }

  function integer() {
    while (has(/\d/)) {
      consume();
    }

    emitToken(INTEGER);
  }

  function identifier() {
    while (has(/[A-Za-z]/)) {
      consume();
    }

    if (tokenSoFar == 'rectangle') {
      emitToken(RECTANGLE);
    } else if (tokenSoFar == 'circle') {
      emitToken(CIRCLE);
    } else if (tokenSoFar == 'repeat') {
      emitToken(REPEAT);
    } else if (tokenSoFar == 'end') {
      emitToken(END);
    } else if (tokenSoFar == 'to') {
      emitToken(TO);
    } else {
      emitToken(IDENTIFIER);
    }
  }
  
  // public stuff
  this.lex = function() {
    iToRead = 0;      // begin at the beginning
    tokenSoFar = '';  //   having seen nothing
    tokens = [];

    while (iToRead < program.length) {
      if (has(/\+/)) {
        consume();
        emitToken(PLUS);
      } else if (has(/[ \t\n\r]/)) {
        consume();
        tokenSoFar = '';
      } else if (has(/\*/)) {
        consume();
        emitToken(TIMES);
      } else if (has(/\)/)) {
        consume();
        emitToken(RIGHT_PARENTHESIS);
      } else if (has(/\(/)) {
        consume();
        emitToken(LEFT_PARENTHESIS);
      } else if (has(/\d/)) {
        integer();
      } else if (has(/[A-Za-z]/)) {
        identifier();
      } else if (has(/-/)) {
        consume();
        if (has(/\d/)) {
          integer();
        } else {
          emitToken(MINUS);
        }
      } else if (has(/=/)) {
        consume();
        emitToken(ASSIGN);
      } else {
        throw 'You fell into the tomb of the unknown token: ' + program.charAt(iToRead);
      }
    }

    emitToken(EOF);

    return tokens;
  }
}

// PARSER ---------------------------------------------------------------------

function Block(statements) {
  this.evaluate = function(scope) {
    statements.forEach(statement => statement.evaluate(scope));
  }
}

function StatementRectangle(left, top, width, height) {
  this.evaluate = function(scope) {
    var l = left.evaluate(scope);
    var t = top.evaluate(scope);
    var w = width.evaluate(scope);
    var h = height.evaluate(scope);
    context.fillStyle = 'rgb(' + scope.red + ',' + scope.green + ',' + scope.blue + ')';
    context.fillRect(l, t, w, h);
  }
}

function StatementCircle(x, y, radius) {
  this.evaluate = function(scope) {
    var xx = x.evaluate(scope);
    var yy = y.evaluate(scope);
    var rr = radius.evaluate(scope);
    context.fillStyle = 'rgb(' + scope.red + ',' + scope.green + ',' + scope.blue + ')';
    context.beginPath();
    context.arc(xx, yy, rr, 0, 2 * Math.PI);
    context.fill();
  }
}

function StatementAssignment(name, rhs) {
  this.evaluate = function(scope) {
    scope[name] = rhs.evaluate(scope);
  }
}

function StatementRepeat(count, body) {
  this.evaluate = function(scope) {
    var n = count.evaluate(scope);
    for (var i = 0; i < n; ++i) {
      body.evaluate(scope);
    }
  }
}

function StatementDefine(name, formals, body) {
  this.evaluate = function(scope) {
    scope[name] = {name: name, formals: formals, body: body};
  }
}

function StatementCall(name, actuals) {
  this.evaluate = function(scope) {
    var f = scope[name]; 
    var innerScope = {};
    for (var i = 0; i < actuals.length; ++i) {
      innerScope[f.formals[i]] = actuals[i].evaluate(scope);
    }
    f.body.evaluate(innerScope);
  }
}

function ExpressionInteger(literal) {
  this.evaluate = function(scope) {
    return literal;
  }
}

function ExpressionVariable(name) {
  this.evaluate = function(scope) {
    return scope[name];
  }
}

function ExpressionAdd(l, r) {
  this.evaluate = function(scope) {
    return l.evaluate(scope) + r.evaluate(scope);
  }
}

function ExpressionSubtract(l, r) {
  this.evaluate = function(scope) {
    return l.evaluate(scope) - r.evaluate(scope);
  }
}

function ExpressionMultiply(l, r) {
  this.evaluate = function(scope) {
    return l.evaluate(scope) * r.evaluate(scope);
  }
}

function ExpressionDivide(l, r) {
  this.evaluate = function(scope) {
    return l.evaluate(scope) / r.evaluate(scope);
  }
}

function Parser(tokens) {
  var iToParse;

  function has(tokenType) {
    return iToParse < tokens.length && tokens[iToParse].type == tokenType;
  }

  function consume() {
    ++iToParse;
  }

  function program() {
    var statements = [];
    while (!has(EOF)) {
      statements.push(statement());
    }
    return new Block(statements);
  }

  function endblock() {
    var statements = [];
    while (!has(END)) {
      statements.push(statement());
    }
    consume(); // eat END
    return new Block(statements);
  }

  function define() {
    consume();
    if (has(IDENTIFIER)) {
      var name = tokens[iToParse].source;
      consume();

      // assert (
      consume();

      var formals = [];
      while (!has(RIGHT_PARENTHESIS)) {
        formals.push(tokens[iToParse].source);
        consume();
      }

      // assert )
      consume();

      var body = endblock();
      return new StatementDefine(name, formals, body);
    } else {
      throw 'Expected function name.';
    }
  }

  function statement() {
    if (has(RECTANGLE)) {
      consume();
      var left = expression();
      var top = expression();
      var width = expression();
      var height = expression();
      return new StatementRectangle(left, top, width, height);
    } else if (has(CIRCLE)) {
      consume();
      var x = expression();
      var y = expression();
      var radius = expression();
      return new StatementCircle(x, y, radius);
    } else if (has(IDENTIFIER)) {
      var name = tokens[iToParse].source;
      consume();
      if (has(ASSIGN)) {
        consume(); // eat =
        var rhs = expression();
        return new StatementAssignment(name, rhs);
      } else if (LEFT_PARENTHESIS) {
        consume();

        var actuals = [];
        while (!has(RIGHT_PARENTHESIS)) {
          actuals.push(expression());
        }

        consume(); // eat )
        return new StatementCall(name, actuals);
      } else {
        throw 'Expected ASSIGN. Unexpected token: ' + tokens[iToParse].source;
      }
    } else if (has(REPEAT)) {
      consume();
      var count = expression();
      var body = endblock();
      return new StatementRepeat(count, body);
    } else if (has(TO)) {
      return define();
    } else {
      throw 'Unexpected token: ' + tokens[iToParse].source;
    }
  }

  function expression() {
    var e = multiplicative();
    while (has(PLUS) || has(MINUS)) {
      var isPlus = has(PLUS);
      consume();
      var f = multiplicative();
      if (isPlus) {
        e = new ExpressionAdd(e, f);
      } else {
        e = new ExpressionSubtract(e, f);
      }
    }
    return e; 
  }

  function multiplicative() {
    var e = atom();
    while (has(TIMES) || has(DIVIDE)) {
      var isTimes = has(TIMES);
      consume();
      var f = atom();
      if (isTimes) {
        e = new ExpressionMultiply(e, f);
      } else {
        e = new ExpressionDivide(e, f);
      }
    }
    return e; 
  }

  function atom() {
    if (has(INTEGER)) {
      var e = new ExpressionInteger(parseInt(tokens[iToParse].source));
      consume();
      return e;
    } else if (has(IDENTIFIER)) {
      var e = new ExpressionVariable(tokens[iToParse].source);
      consume();
      return e;
    } else if (has(LEFT_PARENTHESIS)) {
      consume();
      var e = expression();
      if (has(RIGHT_PARENTHESIS)) {
        consume();
        return e;
      } else {
        throw 'Expected ). Unexpected token: ' + tokens[iToParse].source;
      }
    } else {
      throw 'Expected expression. Unexpected token: ' + tokens[iToParse].source;
    }
  }

  this.parse = function() {
    iToParse = 0;
    return program();
  }
}
